@@ -4,9 +4,6 @@
 
 __version__ = '$Revision: 1.28 $'
 
-import os
-import os.path
-import threading
 import logging
 import logging.handlers
 import sys
@@ -35,4 +32,3 @@ def add_file_handler(filename):
     log.addHandler(handler)
     
 logging.setLoggerClass(LocalLogger)
-
