@@ -5,7 +5,7 @@
 
 """
 This module provides Network Appliance specific classes, most notably
-a NetAppProvisioner that implements a Provisioner for Network Appliance
+a ZAPIProvisioner that implements a Provisioner for Network Appliance
 equipment through the use of the NetApp ZAPI.
 """
 import os
@@ -33,94 +33,6 @@ from change_command import ExpectSet, CommandChange
 from zapi import ZAPITool
 import util
 
-class ZAPIProtocol(client.HTTPPageGetter):
-    """
-    A renamed HTTPPageGetter
-    """
-    quietLoss = 0
-    # We need to authenticate, so override the connectionMade method
-
-    def connectionMade(self):
-        """
-        Custom method for sending a ZAPI request to a filer in the
-        format that it expects.
-        """
-        method = getattr(self.factory, 'method', 'GET')
-        self.sendCommand(method, self.factory.path)
-        self.sendHeader('Host', self.factory.headers.get("host", self.factory.host))
-        self.sendHeader('User-Agent', self.factory.agent)
-        data = getattr(self.factory, 'postdata', None)
-        if data is not None:
-            self.sendHeader("Content-Length", str(len(data)))
-
-        if self.factory.user:
-            auth = base64.encodestring('%s:%s' % (
-                self.factory.user, self.factory.password))
-            self.sendHeader('Authorization', 'Basic %s' % auth)
-            
-        if self.factory.cookies:
-            l=[]
-            for cookie, cookval in self.factory.cookies.items():  
-                l.append('%s=%s' % (cookie, cookval))
-            self.sendHeader('Cookie', '; '.join(l))
-            pass
-
-        self.endHeaders()
-        self.headers = {}
-        
-        if data is not None:
-            self.transport.write(data)
-
-    def endHeaders(self):
-        pass
-
-class ZAPIResult:
-    """
-    An encapsulation of a ZAPI result.
-    """
-    zapi_namespace_id = "http://www.netapp.com/filer/admin"
-
-    re_xmlns = re.compile("(?P<start>.*)\s+xmlns='(?P<namespace>.*)'(?P<end>.*)", re.S) 
-    #re_xmlns = re.compile(r"(.*)xmlns='http://www.netapp.com/filer/admin'(.*)", re.M)
-    
-    def __init__(self, xmlstring):
-
-        self.status = None
-        self.errno = None
-        self.reason = None
-
-        self.rawxml = xmlstring
-
-        # strip off the xmlns bit before parsing
-        # This is done because lxml puts a namespace string at the front of
-        # every tag, and doesn't appear to support a default namespace.
-        m = self.re_xmlns.match(xmlstring)
-        if m:
-            xmlstring = '%s%s' % (m.group('start'), m.group('end') )
-            
-        self.xmlstring = xmlstring
-
-        log.debug("xmlstring is: %s", self.xmlstring)
-        
-        parser = etree.XMLParser(ns_clean=True)
-        self.tree = etree.fromstring(xmlstring, parser)
-        
-        self.results = self.tree.find('results')
-        self.status = self.results.attrib['status']
-        if self.status == 'failed':
-            self.errno = int(self.results.attrib['errno'])
-            self.reason = self.results.attrib['reason']
-        log.debug("zapi status: %s", self.status)
-
-    def __repr__(self):
-        return '<%s: %s, %s>' % (self.__class__, self.status, etree.tostring(self.tree) )
-
-class ZAPIFactory(client.HTTPClientFactory):
-    """
-    A renamed HTTPClientFactory
-    """
-    protocol = ZAPIProtocol
-
 class IZAPIProvisioner(IProvisioner):
     pass
 
@@ -138,17 +50,22 @@ class ZAPIProvisioner(Provisioner):
 
     implements(IZAPIProvisioner)
 
-    def __init__(self, name='', namespace={}, authoritarian=False, autobackout=False, command_timeout=300):
+    def __init__(self, name='', namespace={},
+                 authoritarian=False, autobackout=False,
+                 sessionlog=None,
+                 command_timeout=300):
         """
         Create a new ZAPIProvisioner.
         """
         # FIXME: May not be required
-        Provisioner.__init__(self, name, namespace, authoritarian, autobackout)
+        Provisioner.__init__(self, name, namespace, authoritarian, autobackout,
+                             sessionlog=sessionlog)
         self.command_timeout = 300
 
         # Set up a deferred dictionary, keyed to uniquely identify a set
         # of commands (ExpectSet) running on given device
         self.all_commands_defer = {}
+        self.all_commands_t = {}
 
         # Similarly, key up a series of zapi requests that are outstanding
         # so we can parse the results before returning them via .callback()
@@ -161,6 +78,7 @@ class ZAPIProvisioner(Provisioner):
         Issue a ZAPI request to a device.
         """
         self.authority_check(command)
+        log.debug("issuing command: %s", command)
         return self.zapi_tool.zapi_request(device, command, timeout=timeout)
             
     def zapi_system_command(self, ignored, device, command, timeout=30):
@@ -168,6 +86,7 @@ class ZAPIProvisioner(Provisioner):
         Run a commandline command via ZAPI.
         """
         self.authority_check(command)
+        log.debug("issuing command: %s", command)
         return self.zapi_tool.zapi_system_command(device, command, timeout=timeout)
 
     def zapi_get_file(self, ignored, device, filename, root='/vol/vol0'):
@@ -190,7 +109,7 @@ class ZAPIProvisioner(Provisioner):
         return self.zapi_system_command(device, 'uptime')
         #return self.zapi_request(device, "<system-get-version/>")
 
-    def run_commands(self, ignored, device, expectset, namespace):
+    def run_commands(self, ignored, device, expectset, namespace, delay=0):
 
         self.cmdoutput = ''
         self.exitcode = None
@@ -256,7 +175,9 @@ class ZAPIProvisioner(Provisioner):
                         self.all_commands_defer.errback( e )
                         return self.all_commands_defer[deferred_key]
 
-                    d.addCallback(self.zapi_request, device, cmdstring, self.command_timeout)
+                    # Do substitution on the device with the current namespace
+                    device_addr = util.substituteVariables(str(device), namespace)
+                    d.addCallback(self.zapi_request, device_addr, cmdstring, self.command_timeout)
                     d.addCallbacks(self.command_completed, self.command_failed, callbackArgs=(namespace,), errbackArgs=(namespace,) )
 
                 elif cmdnode.tag == 'getfile':
@@ -316,7 +237,10 @@ class ZAPIProvisioner(Provisioner):
         # Set up callbacks for when all commands are finished.
         # If no commands are run, this will get called with success immediately.
         d.addCallbacks(self.all_commands_done, self.all_commands_failure, callbackArgs=(deferred_key, namespace,), errbackArgs=(deferred_key, namespace,))
-            
+
+        # Set up a timeout to fire if this isn't called back fast enough
+        # FIXME: Use the actual timeout value here
+        self.all_commands_t[deferred_key] = reactor.callLater(30, self.all_commands_timeout, deferred_key)
         return self.all_commands_defer[deferred_key]
 
     def zapi_file_received(self, (results, filedata), device, remotepath, localpath, absolute=False):
@@ -365,13 +289,7 @@ class ZAPIProvisioner(Provisioner):
         """
         If a command failed, pass the failure up the Deferred chain.
         Don't consume the error.
-        FIXME: This should probably be a deferred list or similar.
         """
-        # Don't log here, or the same error will log once for
-        # every command in the chain. This is sortof recurive due
-        # to the way the deferred are chained together above in
-        # the d.addCallback(self.spawn_command) setup
-        
         #self.all_commands_defer.errback( Exception(errorstr) )
         return failure
 
@@ -380,14 +298,22 @@ class ZAPIProvisioner(Provisioner):
         Called when all the commands have completed.
         """
         log.debug("All commands have finished.")
+        self.all_commands_t[deferred_key].cancel()
         self.all_commands_defer[deferred_key].callback( result )
         
     def all_commands_failure(self, failure, deferred_key, namespace):
         """
         Some kind of failure in the commands occurred.
         """
+        self.all_commands_t[deferred_key].cancel()
         self.all_commands_defer[deferred_key].errback( failure )
 
+    def all_commands_timeout(self, deferred_key):
+        """
+        ZAPI commands timed out
+        """
+        self.all_commands_defer[deferred_key].errback( Exception("ZAPI command timed out") )
+        
     def authority_check(self, command):
         """
         Check to see if we should proceed or not
@@ -422,6 +348,7 @@ class ZAPIExpectSet(ExpectSet):
 
         for child in self.known_children:
             for cmd in node.xpath('%s' % child):
+                log.debug("zapicmd: %s", cmd)
                 log.debug("Adding a zapi command to the expectset: %s", etree.tostring(cmd))
                 self.zapi_commands.append(etree.tostring(cmd))
                 pass
