@@ -1,9 +1,13 @@
 #
 # $Id$
 #
-# Command based provisioning, which breaks out into a shell in order
-# to run a command-line that will somehow connect to a remote device
-# and allow shell style commands to be run on it.
+##COPYRIGHT##
+
+"""
+Command based provisioning, which breaks out into a shell in order
+to run a command-line that will somehow connect to a remote device
+and allow shell style commands to be run on it.
+"""
 
 import re
 
@@ -44,14 +48,20 @@ class CommandProvisioner(Provisioner):
     """
     implements(ICommandProvisioner)
     
-    command_re = re.compile(r'^(?P<start>.*)"(?P<quoted>.*)"(?P<end>.*)$')
+    command_re = re.compile(r'^(?P<start>.*)["\'](?P<quoted>.*)["\'](?P<end>.*)$')
 
     def parse_config_node(self, node):
         Provisioner.parse_config_node(self, node)
         pass
 
-    def __init__(self, name='', namespace={}, authoritarian=False, autobackout=False, command_timeout=300):
-        Provisioner.__init__(self, name, namespace, authoritarian, autobackout)
+    def __init__(self, name, namespace={},
+                 authoritarian=False, autobackout=False,
+                 sessionlog=None,
+                 command_timeout=300):
+        Provisioner.__init__(self, name, namespace,
+                             authoritarian=authoritarian,
+                             autobackout=autobackout,
+                             sessionlog=sessionlog)
         self.command_timeout = int(command_timeout)
 
     def split_command(self, cmdstring):
@@ -73,17 +83,16 @@ class CommandProvisioner(Provisioner):
 
         return cmdstring.split()
 
-    def run_command(self, ignored, cmdstring):
+    def run_commands(self, ignored, device, expectset, namespace):
         """
-        Execute a command on the element, returning the results of the command.
+        Execute an expectset of commands on the device, returning the results of the command.
 
-        @returns: (return_code, stdout, stderr) where
+        @returns: (return_code, results) where
           return_code: is the exit code of the command, as an integer
-          stdout: is a string containing anything printed to STDOUT
-          stderr: is a string containing anything printed to STDERR
+          results: is a string containing anything printed to the subprocess
         """
-        log.debug("Executing provisioning command: %s" % cmdstring)
-        return defer.succeed( (0, '', '') )
+        log.debug("Executing expectset: %s" % expectset)
+        return defer.succeed( CommandChangeResult((0, '')) )
 
 class CommandChangeResult(ChangeResult):
 
@@ -94,18 +103,32 @@ class CommandChangeResult(ChangeResult):
         self.cmdoutput = result[1]
 
     def __str__(self):
-        return self.cmdoutput
+        return "%s: %s" % (self.exitcode, self.cmdoutput)
 
+    def __repr__(self):
+        return "%s: exitcode: %s" % (self.__class__.__name__, self.exitcode)
+
+    
 class ConnectingProvisioner(CommandProvisioner):
     """
     A ConnectingProvisioner runs a command to connect to the remote device
     and then runs a series of commands by sending them to the remote
     device via STDIN, and receives the output from STDOUT and STDERR.
     """
-    def parse_config_node(self, node):
-        CommandProvisioner.parse_config_node(self, node)
+    required_tags = CommandProvisioner.required_tags + [ 'connect_command', ]
+    optional_tags = CommandProvisioner.optional_tags + []
 
-    def connect(self, device):
+    def parse_config_node(self, node):
+        """
+        Deals with additional configuration that I may be provided with.
+        """
+        CommandProvisioner.parse_config_node(self, node)
+        # parse command definition
+        if node.tag == 'connect_command':
+            self.conn_cmd = str(node.text)
+            log.debug("Set provisioner connecting command to: %s", self.conn_cmd)
+
+    def connect(self, device, namespace):
         self.connectedDevice = device
         
         self.connecting_d = defer.Deferred()
@@ -113,18 +136,20 @@ class ConnectingProvisioner(CommandProvisioner):
 
         log.debug("connecting to device '%s'", device)
 
-        cmd = [ '/usr/bin/ssh',
-                '-T',
-                '-o','BatchMode=yes',
-                '-i','/home/daedalus/.ssh/localconnect',
-                'localhost' ]
-        #cmd = [ '/usr/bin/ssh', 'localhost' ]
-        self.ep = ExpectProtocol(self)
-        self.p = reactor.spawnProcess(self.ep, cmd[0], cmd )
+        # Do variable substitution on the connect command
+        cmdstring = str(util.substituteVariables(self.conn_cmd, namespace))
+        cmd = self.split_command(cmdstring)
+        
+#         cmd = [ '/usr/bin/ssh',
+#                 '-T',
+#                 '-o','BatchMode=yes',
+#                 '-i','/home/daedalus/.ssh/localconnect',
+#                 'localhost' ]
+        self.ep = ExpectProtocol(self, self.sessionlog)
+        log.debug("spawning process: %s", cmd)
+        self.p = reactor.spawnProcess(self.ep, cmd[0], cmd, usePTY=True )
 
-        #log.debug("timeout thing: %s" % self.childConnectingTimeout)
-
-        return self.childConnectingDefer
+        return self.connecting_d
 
     def connection_timeout(self, device):
         """
@@ -142,28 +167,59 @@ class ConnectingProvisioner(CommandProvisioner):
         completed successfully.
         """
         log.debug("successfully connected to child")
-        self.childConnectingTimeout.cancel()
+        self.connection_timeout_d.cancel()
         log.debug("sent timeout cancel")
-        self.childConnectingDefer.callback(self)
+        self.connecting_d.callback(self)
 
     def childConnectFailure(self):
         """
-        This gets called by the child when it finishes connecting.
-        Cancel the timeout and notify waiting things that the connection
-        completed successfully.
+        This gets called when the connection fails
         """
         log.error("connection by child failed")
-        self.connection_timeout_d.cancel()
+        #self.connection_timeout_d.cancel()
         self.connecting_d.errback(ValueError("Child connection failure!"))
 
-    def run_commands(self, ignored, device, expectset, namespace):
+    def childCommandFailure(self, error):
+        """
+        Called when the child command fails for some reason that
+        is probably related to the command used for connecting.
+        """
+        self.connecting_d.errback(error)        
+
+    def run_commands(self, ignored, device, expectset, namespace, delay=0):
+        """
+        Delay is optional. Used to run the command after a delay.
+        """
         log.debug('running commands on remote host: %s' % expectset)
 
-        return self.ep.run_remote_commands(expectset)
+        # Check to see if I'm connected. If not, connect first.
+        if getattr(self, 'ep', None) is None:
+            def connected(ignored, d):
+                log.info("unconnected delayed command")
+                reactor.callLater(delay, self.run_delayed_command, d, expectset, namespace)
+                #return self.ep.run_remote_commands(expectset, namespace)
+            d = self.connect(device, namespace)
+            d.addCallback(connected, d)
+            return d
+
+        else:
+            d = defer.Deferred()
+            log.info("connected delayed command")
+            reactor.callLater(delay, self.run_delayed_command, d, expectset, namespace)
+            #return self.ep.run_remote_commands(expectset, namespace)
+            return d
 
         # Run the command using the ExpectProtocol
         #return defer.succeed( (0, '', '') )
 
+    def run_delayed_command(self, d, expectset, namespace):
+        """
+        Runs a possibly delayed command, and chains the callbacks
+        to the original Deferred.
+        """
+        newd = self.ep.run_remote_commands(expectset, namespace)
+        newd.chainDeferred(d)
+        
     def disconnect(self, ignored, device):
         """
         Called when the provisioner is no longer required.
@@ -178,6 +234,21 @@ class ConnectingProvisioner(CommandProvisioner):
         log.error("Command failed: %s", errorstr)
         self.all_commands_defer[deferred_key].errback( failure )
 
+    def change_failure(self, failure, change, namespace):
+        """
+        Override base class in order to disconnect when the change is
+        finished.
+        """
+        # disconnect from the remote device
+        if getattr(self, 'ep', None) is not None:
+            self.ep.transport.loseConnection()
+        Provisioner.change_failure(self, failure, change, namespace)
+
+    def change_complete_success(self, result, change, namespace):
+        # disconnect from the remote device
+        self.ep.transport.loseConnection()
+        Provisioner.change_complete_success(self, result, change, namespace)
+    
 class ExpectProtocol(protocol.Protocol):
     """
     The ExpectProtocol allows you to supply a series of
@@ -203,15 +274,20 @@ class ExpectProtocol(protocol.Protocol):
     being populated from the previous expect re.match operation.
     """
 
-    def __init__(self, provisioner):
+    def __init__(self, provisioner, sessionlog):
         
         self.parent = provisioner
+        self.sessionlog = sessionlog
         self.saidHello = False
         self.waitingForCommand = None
         self.databuf = ''
         self.data_wait_timeout = None
 
-        self.results = ''
+        self.cmdoutput = ''
+        self.expectset = None
+        # Commands are assumed to succeed unless the expect string
+        # cannot be found.
+        self.exitcode = 0
 
     def connectionMade(self):
         """
@@ -227,108 +303,200 @@ class ExpectProtocol(protocol.Protocol):
     def childDataReceived(self, fd, data):
         log.debug("received data from child process %s: >%s<" % (fd, data) )
 
-        try:
-            self.databuf += data
-            log.debug("databuf: >%s<" % self.databuf)
-            self.do_expect_processing()
-
-        # force any unexpected errors to call the errback
-        except Exception, e:
-            self.waitingForCommand.errback(e)
-
+        self.got_data(data)
         # ignore stuff from STDERR?
         #if fd == 2:
         #    pass
 
+    def outReceived(self, data):
+        """
+        Alternate version of childDataReceived() for PTYProcess, rather
+        than something that allows full control of child file descriptors.
+        """
+        self.got_data(data)
+        
+    def got_data(self, data):
+        """
+        Generic function used when data is received from the child process,
+        whatever kind of subprocess it is.
+        """
+        #log.debug("data: %s" % data)
+        self.databuf += data
+        if self.sessionlog is not None:
+            self.sessionlog.write(data)
+            self.sessionlog.flush()            
+            pass
+        #log.debug("databuf: >%s<" % self.databuf)
+        self.do_expect_processing()
+
+#         # force any unexpected errors to call the errback
+#         except Exception, e:
+#             try:
+#                 log.error("Exception in childDataReceived(): %s: %s", e.__class__, e)
+#                 self.waitingForCommand.errback(e)
+#             except defer.AlreadyCalledError:
+#                 pass
+            
     def processEnded(self, status_object):
         log.debug("process ended: %s" % status_object)
 
         if isinstance(status_object.value, ProcessDone):
             log.debug("process exited ok")
+            self.exitcode = 0
 
         elif isinstance(status_object.value, ProcessTerminated):
-            log.debug("process did not exit ok: %s" % status_object.value.exitCode)
-            self.parent.childConnectFailure()
-
-        log.debug("handled error")
+            errmsg = "process did not exit ok: %s" % status_object.value.exitCode
+            log.debug(errmsg)
+            self.exitcode = status_object.value.exitCode
+            #self.parent.childCommandFailure(ValueError(errmsg))
 
     def connectionFinished(self, status):
         self.parent.childConnected()
 
-    def run_remote_commands(self, expectset):
+    def processExited(self, reason):
+        pass
+
+    def run_remote_commands(self, expectset, namespace):
         """
         Run a set of commands, using expect processing to do it.
         """
-        self.expectset = expectset
+        self.namespace = namespace
+        
+        log.debug("running remote commands: %s", expectset)
+        self.expectset = expectset.copy()
         self.waitingForCommand = defer.Deferred()
         self.do_expect_processing()
         return self.waitingForCommand
 
     def do_expect_processing(self):
 
-        if self.data_wait_timeout:
-            self.data_wait_timeout.cancel()
+        if self.expectset is None:
+            return
 
-        # get the first expression/cmd pair from the expect commands
-        expr, cmdstring = self.expectset.commands[0]
-        self.expectset = self.expectset.commands[1:]
-        
-        # check the expr to see if we should process it
-        if expr is None:
-            if cmdstring is not None:
-                log.debug("No expression to wait for. Running command '%s' immediately." % cmdstring)
-                self.issue_command(cmdstring)
-            else:
-                log.debug("cmdstring is 'None', so nothing to do")
+        #log.debug("doing expect processing: %s", self.expectset.commands)
+        # loop over the commands, breaking out if we need to
+        while 1:
+            try:
+                log.debug("fetching next expectset tuple...")
+                expr, cmdstring, timeout = self.expectset.pop_command()
+                if timeout == -1:
+                    timeout = self.parent.command_timeout
 
-        else:
-            log.debug("checking for '%s' in databuf: %s", expr, self.databuf)
-            re_expr = re.compile(expr)
-            log.debug("compiled expr")
+                log.debug("Seeking: %s", expr)
 
-            # Match in multi-line mode, and '.' character matches \n
-            match = re_expr.search(self.databuf)
-            log.debug("match object is: %s", match)
-            if match:
-                # check the current recv buffer to see if the expect string is a match
-                log.debug("Found expr '%s' in databuf! Yay!" % expr)
-
-                # reset the databuffer whenever we match something
-                # FIXME: maybe not do this?
-                self.results += self.databuf
+            except IndexError:
+                log.debug("No more commands to run.")
+                self.cmdoutput += self.databuf
                 self.databuf = ''
-                
-                if cmdstring is not None:
-                    self.issue_command(cmdstring)
-                    pass
-                else:
-                    # If the command is None, this is the end of processing,
-                    # so shut down the connection and trigger command success processing.
-                    self.transport.loseConnection()
-                    self.waitingForCommand.callback(self.results)
-                    
-            else:
-                log.debug("cannot find '%s' in databuf. Waiting for more data..." % expr)
-                self.expectset.insert(0, (expr, cmdstring) )
-                self.data_wait_timeout = reactor.callLater(10, self.data_wait_too_long)
-                pass
-            pass
 
+                # This can sometime happen when extra, spurious data is
+                # received after we've finished what we wanted to do.
+                # We just ignore that.
+                try:
+                    self.waitingForCommand.callback( CommandChangeResult((self.exitcode, self.cmdoutput)))
+                except defer.AlreadyCalledError:
+                    pass
+
+                # All done, so exit out of the loop
+                return
+            # See if we need to wait for an expression
+            if expr is None:
+                self.process_command(cmdstring)
+            else:
+                # There is an expression, so process it.
+                expr = str(util.substituteVariables(expr, self.namespace))
+                if not self.find_expect(expr, cmdstring, timeout):
+                    break
+
+    def process_command(self, cmdstring):
+        """
+        Process the command associated with an expect.
+        """
+        if cmdstring is None:
+            log.debug("no command. nothing to do")
+            # No command, so nothing to do.
+            pass
+        else:
+            try:
+                cmdstring = str(util.substituteVariables(cmdstring, self.namespace))
+                log.debug("No expression to wait for. Running command '%s' immediately." % cmdstring)
+            except Exception, e:
+                self.waitingForCommand.errback(e)
+            self.issue_command(cmdstring)
+            
+    def find_expect(self, expr, cmdstring, timeout):
+        """
+        See if we can find the expect expression.
+        """
+        #log.debug("checking for '%s' in databuf[len %d]: %s", expr, len(self.databuf), self.databuf)
+        if len(self.databuf) == 0:
+            # No data yet, so wait
+            self.wait_for_expect(expr, cmdstring, timeout)
+            return False
+
+        # Match in multi-line mode, and '.' character matches \n
+        re_expr = re.compile(expr)
+        match = re_expr.search(self.databuf)
+        log.debug("match object is: %s", match)
+        if match:
+            # check the current recv buffer to see if the expect string is a match
+            log.debug("Found expr '%s' in databuf! Yay!" % expr)
+
+            if getattr(self, 'data_wait_timeout', None) is not None:
+                log.debug("Found expected string. cancelling timeout.")
+                self.data_wait_timeout.cancel()
+                del self.data_wait_timeout
+                pass
+
+            # reset the databuffer whenever we match something to everything
+            # that comes after the matched part.
+            log.debug("Match end is at: %d", match.end())
+            self.cmdoutput += self.databuf[:match.end()]
+            self.databuf = self.databuf[match.end():]
+
+            self.process_command(cmdstring)
+
+            # We found something, so let caller know
+            return True
+
+        # The expression wasn't found, so we put the command back on the
+        # stack and wait for more data to trigger this processing again.
+        else:
+            self.wait_for_expect(expr, cmdstring, timeout)
+            return False
+
+    def wait_for_expect(self, expr, cmdstring, timeout):
+        """
+        Can't find the expect expression yet, so wait for it
+        until a timeout.
+        """
+        log.debug("cannot find '%s' in databuf. Waiting for more data..." % expr)
+        self.expectset.push_command(expr, cmdstring, timeout)
+
+        # Set a timeout. If the timeout value is 0, disable the timeout.
+        if timeout > 0:
+            if getattr(self, 'data_wait_timeout', None) is None:
+                log.debug("Setting timeout value to: %d", timeout)
+                self.data_wait_timeout = reactor.callLater(timeout, self.data_wait_too_long)
+            pass
+            
     def issue_command(self, cmdstring):
         """
         Write the command to the remote device.
         """
         self.parent.check_authoritarian()
         
-        log.debug("running command: %s" % cmdstring)                    
-        self.transport.writeToChild(0, '%s\n' % cmdstring)
-        self.data_wait_timeout = reactor.callLater(10, self.data_wait_too_long)
-
+        log.debug("running command: %s" % cmdstring)
+        self.transport.writeSomeData('%s\n' % cmdstring)
+        # This may well get recorded anyway if the session echos
+#         if self.sessionlog is not None:
+#             self.sessionlog.write('%s\n' % cmdstring)
+#             self.sessionlog.flush()
     def data_wait_too_long(self):
         """
         We have waited for data for too long, so return an error
         """
-        self.waitingForCommand.errback( ValueError("Timed out waiting for command output.") )
+        self.waitingForCommand.errback( ValueError("Timed out waiting for expect string.") )
 
 class MultiConnectingProvisioner(ConnectingProvisioner):
     """
@@ -394,7 +562,7 @@ class MultiConnectingProvisioner(ConnectingProvisioner):
         self.all_commands_defer[deferred_key] = defer.Deferred()
 
         d = defer.succeed(None)
-        for (expr, cmdstring) in expectset.commands:
+        for (expr, cmdstring, timeout) in expectset.commands:
 
             try:
                 # Perform variable substitution on the command string
@@ -522,9 +690,8 @@ class SingleCommandProtocol(protocol.Protocol):
 
     def childDataReceived(self, fd, data):
         #log.debug("received data from child process %s: >%s<" % (fd, data) )
-
         self.databuf += data
-        log.debug("databuf: >%s<" % self.databuf)
+        #log.debug("databuf: >%s<" % self.databuf)
 
     def processEnded(self, status_object):
         log.debug("process ended: %s" % status_object)
